// -*- c++ -*-
//------------------------------------------------------------------------------
//                              Config.h
//------------------------------------------------------------------------------
// $Id: Config.h,v 1.18 2004/06/27 01:27:31 vlg Exp $
//------------------------------------------------------------------------------
//  Copyright (c) 2004 by Vladislav Grinchenko
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version
//  2 of the License, or (at your option) any later version.
//------------------------------------------------------------------------------
// 01/03/2004 VLG  Created
//------------------------------------------------------------------------------
#ifndef CONFIG_H
#define CONFIG_H

#include <string>
#include <deque>
using std::deque;
using std::string;

#include <pangomm/fontdescription.h>

#include <assa/Singleton.h>
#include <assa/TimeVal.h>
#include <assa/IniFile.h>
class ASSA::IniFile;

#include "Granule-main.h"

// Record when to review cards in box I (I=DAYS/WEEKS)
//
struct SchedReview 
{
	string m_days;			// number of days
	string m_weeks;			// number of weeks
	ASSA::TimeVal m_tv;		// time in seconds

	string days  () const           { return m_days;  }
	string weeks () const           { return m_weeks; }
	void   days  (const string& d_) { m_days  = d_;   }
	void   weeks (const string& w_) { m_weeks = w_;   }

	void set_delay (const ASSA::TimeVal& load_time_,
					int secs_in_day_, 
					int secs_in_week_);
	void dump (int idx_);
};

class Config : public ASSA::Singleton<Config>
{
public:
    typedef deque<string> DHList; // document history list
    typedef DHList::iterator DHList_Iter;
    typedef DHList::const_iterator DHList_CIter;

public:
    Config ();
    ~Config ();

    void save_config ();
    void load_config (int secs_in_day_, int secs_in_week_);
    void dont_save_project () { m_dont_save_project = true; }

    // Accessors
    const string& get_proj_name    () const { return m_proj_name;    }
    const string& get_proj_path    () const { return m_path_name;    }
	const string& get_filesel_path () const { return m_filesel_path; }
    void set_proj_name    (const string& fname_);
	void set_filesel_path (const string& fspath_);

	/// Handling card duplicates
	bool get_remove_duplicates () const  { return m_remove_dups; }
	void set_remove_duplicates (bool b_) { m_remove_dups = b_;   }

	/** Returns an expiration date that should be assigned to a 
		card as it is placed in CardDeck # idx_. 

		@param idx_ CardDeck number.
		@return Card expiration time.
	*/
	ASSA::TimeVal get_expiration (int idx_);

	/** Access to the scheduling settings by index.
	 */
	string get_sched_review_days  (int idx_) { return m_sched [idx_].days  (); }
	string get_sched_review_weeks (int idx_) { return m_sched [idx_].weeks (); }

	void set_sched_review_days  (int idx_, const string& s_) { 
		m_sched [idx_].days (s_); 
	}
	void set_sched_review_weeks (int idx_, const string& s_) { 
		m_sched [idx_].weeks (s_); 
	}

	/// Return timestamp when the program was started.
	ASSA::TimeVal get_load_time () const { return m_load_time; }

	/// Get Real User Name
	const char* get_user_name        () const;

	/** Check Deck side recorded. If Deck is not found, return FRONT
	 */
	SideSelection get_flipside_history (const string& fname_);
	void set_flipside_history (const string& fname_, SideSelection side_);

	/** Text fonts settings
	 */
	Pango::FontDescription& get_font () { return m_font_desc; }
	void set_font (const Glib::ustring& font_);

	/** Setting for sound playing
	 */
	const char* get_snd_player_cmd   () const;
	const char* get_snd_player_args  () const;
	const char* get_snd_archive_path () const;

	void set_snd_player_cmd   (const string& s_) { m_snd_player_cmd = s_;   }
	void set_snd_player_args  (const string& s_) { m_snd_player_args = s_;  }
	void set_snd_archive_path (const string& s_) { m_snd_archive_path = s_; }

	// History-related functions
    string        add_document_history    ();
	void          remove_document_history (const string& fullpath_);
    const DHList& get_document_history    () const { return m_history; }

    /// Dump object's state
    void dump () const;
    void dump_document_history () const;

private:
    Config (const Config&);		// copy constructor
    Config& operator= (const Config&); // copy assignment: cleanup & copy

	typedef std::pair<string, SideSelection> sidesel_t;
    typedef std::vector<sidesel_t> flipside_t;
    typedef flipside_t::iterator flipside_iter_t;

private:
    string m_proj_name;			        // Project name
    string m_path_name;			        // Path to configuration file
	string m_filesel_path;		        // Path of the last FileSelection dialog
    bool   m_dont_save_project;	        // Don't save project file
	Pango::FontDescription m_font_desc;	// Default font
	string m_snd_player_cmd;	        // Sound player string
	string m_snd_player_args;           // Sound player arguments formatted 
                                        // string.
	string m_snd_archive_path;			// Path to dictionary sound archive
	string m_user_name;					// Real user name from /etc/passwd

	DHList         m_history;	        // CardFile history list.
	ASSA::IniFile* m_config;
	flipside_t     m_flipside_history;	// Which side to show for each Deck.
	bool           m_remove_dups;       // Purge CardDeck of duplicate VCards
	ASSA::TimeVal  m_load_time;	        // Time the program has been loaded
	SchedReview    m_sched [CARD_BOX_SIZE];	// Review schedule
	int            m_secs_in_day;       // Seconds in a day
	int            m_secs_in_week;      // Seconds in a week
};

// Global locator

#define CONFIG Config::getInstance()

//------------------------------------------------------------------------------
// Inlines
//------------------------------------------------------------------------------
//
inline void Config::set_font (const Glib::ustring& font_)
{
	m_font_desc = Pango::FontDescription (font_);
}

inline const char* Config::get_snd_player_cmd () const 
{ 
	return m_snd_player_cmd.c_str ();
}

inline const char* Config::get_snd_player_args() const 
{ 
	return m_snd_player_args.c_str();
}

inline const char* Config::get_snd_archive_path () const 
{ 
	return m_snd_archive_path.c_str (); 
}

inline const char* Config::get_user_name () const 
{ 
	return m_user_name.c_str (); 
}

inline ASSA::TimeVal Config::get_expiration (int idx_)
{
	return (m_sched [idx_].m_tv);
}
    
#endif /* CONFIG_H */
