// -*- c++ -*-
//------------------------------------------------------------------------------
//                              Deck.cpp
//------------------------------------------------------------------------------
// $Id: Deck.cpp,v 1.26 2004/07/06 04:34:23 vlg Exp $
//
//  Copyright (c) 2004 by Vladislav Grinchenko
//
//  This program is free software; you can redistribute it and/or 
//  modify it under the terms of the GNU General Public License   
//  as published by the Free Software Foundation; either version  
//  2 of the License, or (at your option) any later version.      
//
// Date: Jan 8, 2004
//------------------------------------------------------------------------------

#include <fstream>
using std::ofstream;

#include <assa/CommonUtils.h>
#include <assa/Fork.h>
#include <assa/TimeVal.h>

#include "Granule.h"
#include "Granule-main.h"
#include "DeckManager.h"
#include "Deck.h"
#include "Card.h"
#include "MainWindow.h"
#include "Config.h"

/** Compare real cards by their IDs
 */
class Compare_Cards {
public:
	int operator () (VCard* const& vc1_, VCard* const& vc2_)
		{
			Card* cr1 = dynamic_cast<Card*> (vc1_);
			Card* cr2 = dynamic_cast<Card*> (vc2_);
			Assert_exit (cr1 && cr2);
			return (cr1->get_id () < cr2->get_id ());
		}
};	

/** C'tor for the new deck
 */
Deck::
Deck (const string& fname_, DeckEditStatus status_) 
	:
	m_fname  (fname_), 
	m_status (status_),
	m_needs_renaming ((status_ == NEW_DECK))
{
	trace_with_mask("Deck::Deck(untitled)",GUITRACE|DECK);

	m_author   = CONFIG->get_user_name ();
	m_snd_path = CONFIG->get_snd_archive_path ();
	{
		ASSA::TimeVal today = ASSA::TimeVal::gettimeofday ();
		m_desc     = "Created on " + today.fmtString ("%c");
	}
	dump_deck_status ();
}

void 
Deck::
load (const string& fname_, DeckAction action_)
{
	trace_with_mask("Deck::load",GUITRACE|DECK);

	DL((DECK|APP,"Loading \"%s\" ...\n", fname_.c_str ()));

	/** Load the Deck in XML format 
	 */
	try {
		xmlpp::DomParser parser;
		parser.set_validate ();
		parser.parse_file (fname_);
		if (!parser) {
			DL((GRAPP,"Failed to parse \"%s\"\n", fname_.c_str ()));
			DECKMGR->on_deck_loaded (this, LOAD_FAILED);
			return;
		}
		if (action_ == INSTALL_DECK) {
			m_fname = fname_;
		}

		/** Root node is managed by the parser
		 */
		DL((GRAPP,"Parsing the deck ...\n"));
		const xmlpp::Node* root_node = parser.get_document ()->get_root_node ();
		Assert_exit (root_node);

		parse_xml_tree (root_node);
		if (action_ == INSTALL_DECK) {
			DECKMGR->on_deck_loaded (this, LOAD_OK);
		}
		m_status = CLEAN_DECK;
		dump_deck_status ();
	}
	catch (const std::exception& ex_) {
		DECKMGR->on_deck_loaded (this, LOAD_FAILED);
	}
}

void 
Deck::
insert (const string& from_fname_)
{
	trace_with_mask("Deck::insert",GUITRACE|DECK);

	load (from_fname_, ADD_TO_DECK); 
	m_status = MODIFIED_DECK;

	/**
	 * Wack duplicates
	 */
	Card* cr_p;
	Card* cr_q;

	std::sort (m_cards.begin (), m_cards.end (), Compare_Cards ());
	cardlist_iterator p = m_cards.begin ();
	cardlist_iterator q = p + 1;
	while (q != m_cards.end ()) {
		cr_p = dynamic_cast<Card*> (*p);
		cr_q = dynamic_cast<Card*> (*q);
		Assert_exit (cr_p && cr_q);
		if (cr_p->get_id () == cr_q->get_id ()) {
			delete (*q);
			m_cards.erase (q);
			q = p;
		}
		else {
			p = q;
		}
		q++;
	}
}

void
Deck::
parse_xml_tree (const xmlpp::Node* node_)
{
	trace_with_mask("Deck::traverse_xml_tree",GUITRACE|DECK);

	char xpath [64];
	const xmlpp::TextNode* text_node;
	const xmlpp::Attribute* attr_node;
	string id;
	xmlpp::NodeSet nset;
	xmlpp::NodeSet::iterator nset_iter;
	Card* card;

	nset = node_->find ("/deck/author/text()");
	Assert_exit ((nset_iter = nset.begin ()) != nset.end ());
	text_node = dynamic_cast<const xmlpp::TextNode*> (*nset_iter);
	Assert_exit (text_node);
	m_author = text_node->get_content ();

	nset = node_->find ("/deck/description/text()");
	Assert_exit ((nset_iter = nset.begin ()) != nset.end ());
	text_node = dynamic_cast<const xmlpp::TextNode*> (*nset_iter);
	Assert_exit (text_node);
	m_desc = text_node->get_content ();

	nset = node_->find ("/deck/sound_path/text()");
	Assert_exit ((nset_iter = nset.begin ()) != nset.end ());
	text_node = dynamic_cast<const xmlpp::TextNode*> (*nset_iter);
	Assert_exit (text_node);
	m_snd_path  = Glib::locale_from_utf8 (text_node->get_content ());

	struct stat st;
	if (::stat (m_snd_path.c_str (), &st) != 0 || !S_ISDIR (st.st_mode)) {
		DL((APP,"Running with sound disabled\n"));
		DL((APP,"Path \"%s\" is wrong/non-existed\n",m_snd_path.c_str ())); 
		m_with_snd = false;
	}
	dump ();

	for (int cards_counter = 1; ; cards_counter++) {
		sprintf (xpath, "/deck/card[%d]", cards_counter);
		nset = node_->find (xpath);
		if ((nset_iter = nset.begin ()) == nset.end ()) {
			break;
		}
		card = new Card;

		sprintf (xpath, "/deck/card[%d]/@id", cards_counter);
		nset = node_->find (xpath);
		Assert_exit ((nset_iter = nset.begin ()) != nset.end ());
		attr_node = dynamic_cast<const xmlpp::Attribute*> (*nset_iter);
		Assert_exit (attr_node);
		id = (attr_node->get_value ()).substr (1);
		card->set_id (::atol (id.c_str ()));

		sprintf (xpath, "/deck/card[%d]/front/text()", cards_counter);
		nset = node_->find (xpath);
		Assert_exit ((nset_iter = nset.begin ()) != nset.end ());
		text_node = dynamic_cast<const xmlpp::TextNode*> (*nset_iter);
		Assert_exit (text_node);
		card->set_question (text_node->get_content ());
		
		sprintf (xpath, "/deck/card[%d]/back/text()", cards_counter);
		nset = node_->find (xpath);
		Assert_exit ((nset_iter = nset.begin ()) != nset.end ());
		text_node = dynamic_cast<const xmlpp::TextNode*> (*nset_iter);
		Assert_exit (text_node);
		card->set_answer (text_node->get_content ());

		sprintf (xpath, "/deck/card[%d]/back_example/text()", cards_counter);
		nset = node_->find (xpath);
		Assert_exit ((nset_iter = nset.begin ()) != nset.end ());
		text_node = dynamic_cast<const xmlpp::TextNode*> (*nset_iter);
		Assert_exit (text_node);
		card->set_example (text_node->get_content ());

		card->mark_clean ();
		m_cards.push_back (card);
		card->dump ();
	}
	DL((ASSA::APP,"Processed %d cards\n", m_cards.size ()));
}

bool
Deck::
erase (VCard* vcard_)
{
	trace_with_mask("Deck::erase",GUITRACE|DECK);

	cardlist_iterator pos;
	Assert_exit (vcard_);

	pos = find (m_cards.begin (), m_cards.end (), vcard_);
	if (pos != m_cards.end ()) {
		m_cards.erase (pos);
		m_status = MODIFIED_DECK;
		dump_deck_status ();
		return true;
	}
	DL((GRAPP, "Card (%d) cannot be deleted (not found)\n", vcard_->get_id ()));
	return false;
}

int
Deck::
save_as (const string& fname_)
{
	trace_with_mask("Deck::save_as",GUITRACE|DECK);
	string fname_ext = fname_;

	/** If extension is missing, add it.
	 */
	string ext (DECK_FILE_EXT);
    if (fname_ext.find (ext) == string::npos) {
		fname_ext += ext;
    }
	if (Glib::file_test (fname_ext, Glib::FILE_TEST_IS_REGULAR)) {
		string msg = "File \n" + fname_ext
			+ "\nalready exists.\nWould you like to delete it first?";
		Gtk::MessageDialog qm (*MAINWIN, msg, false,
							   Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_NONE,
							   true);
		qm.add_button ("Yes", Gtk::RESPONSE_YES);
		qm.add_button ("No", Gtk::RESPONSE_NO);
		if (qm.run () == Gtk::RESPONSE_NO) {
			return -1;
		}
		::unlink (fname_ext.c_str ());
	}

	if (needs_renaming ()) {
		m_fname = fname_ext;
		save ();
		CONFIG->set_flipside_history (m_fname, FRONT);
	}
	else {
		string tmp = m_fname;
		m_fname = fname_ext;
		save ();
		m_fname = tmp;
	}
	return 0;
}

int
Deck::
save ()
{
	trace_with_mask("Deck::save",GUITRACE|DECK);

	xmlpp::Document deck_xml;
	xmlpp::Element* deck_node;
	xmlpp::Element* card_node;
	xmlpp::Element* element;
	xmlpp::TextNode* txt_node;

	::unlink (m_fname.c_str ());
	deck_xml.set_internal_subset ("deck", "", GRAPPXMLDIR "/granule.dtd");
	deck_node = deck_xml.create_root_node ("deck");
	
/** FIXME : can you skip txt_node and do smthing like:

    element->add_content ()->set_content ("blahh...") ?
*/
	element = deck_node->add_child ("author");
	txt_node = element->add_content ();
	txt_node->set_content (m_author);

	element = deck_node->add_child ("description");
	txt_node = element->add_content ();
	txt_node->set_content (m_desc);

	element = deck_node->add_child ("sound_path");
	txt_node = element->add_content ();
	txt_node->set_content (Glib::locale_to_utf8 (m_snd_path));

	cardlist_const_iterator citer = m_cards.begin ();
	while (citer != m_cards.end ()) {
		card_node = deck_node->add_child ("card");
		card_node->set_attribute ("id", (*citer)->get_id_str ());

		element = card_node->add_child ("front");
		txt_node = element->add_content ();
		txt_node->set_content ((*citer)->get_question ());
		
		element = card_node->add_child ("back");
		txt_node = element->add_content ();
		txt_node->set_content ((*citer)->get_answer ());
		
		element = card_node->add_child ("back_example");
		txt_node = element->add_content ();
		txt_node->set_content ((*citer)->get_example ());
		
		citer++;
	}

	deck_xml.write_to_file_formatted (m_fname, "UTF-8");
	m_status = CLEAN_DECK;
	m_needs_renaming = false;

	/** clear up all cards
	 */
	cardlist_iterator iter = m_cards.begin ();
	while (iter != m_cards.end ()) {
		(*iter)->mark_clean ();
		iter++;
	}

	dump_deck_status ();
}

void
Deck::
get_progress (float& pct_done_, std::string& msg_, int idx_)
{
	trace_with_mask("Deck::get_progress",GUITRACE|DECK);

	std::ostringstream os;
	os << (idx_ + 1) << "/" << m_cards.size ();
	msg_ = os.str ();
	if (m_cards.size () <= 1) {
		pct_done_ = 0;
	}
	else {
		pct_done_ = idx_ * 1.0 / (m_cards.size () - 1);
	}
}

VCard* 
Deck::
find_card_by_id (long id_)
{
	trace_with_mask("Deck::find_card_by_id",GUITRACE|DECK);

	cardlist_iterator iter = m_cards.begin ();
	while (iter != m_cards.end ()) {
		if ((*iter)->get_id () == id_) {
			return (*iter);
		}
		iter++;
	}
	return (NULL);
}

/** Rules for playable name are as follows:

    1) A word can be preceeded by "to " to indicate its a verb.
	2) A multiword expressions are ignored.
	3) The WAV file is assumed to be stored in the directory
       with the name of its first letter, i.e. "granule" will be
       stored in the directory "g/", and thus the full name
	   will be "g/granule.wav".

	RETURN: playable name on success; an empty string on error.
*/
void
Deck::
play_card (Deck::cardlist_iterator& iter_)
{
	trace_with_mask("Deck::play_card",GUITRACE|DECK);

	if (! m_with_snd) {
		return;
	}
	std::vector<string> v;
	string w = Glib::locale_from_utf8 ((*iter_)->get_question ());
	ASSA::Utils::split (w.c_str (), v);
	if (v.size () > 1) {
		if (v[0] != "to") {
			return;
		}
		w = v[1];
	}
	string sfn (m_snd_path + "/");
	sfn += w[0];
	sfn += "/" + w + ".wav";

	struct stat st;
	if (::stat (sfn.c_str (), &st) != 0 || !S_ISREG (st.st_mode)) {
		DL((APP,"Can't stat sound file \"%s\"\n", sfn.c_str ()));
		return;
	}
	char buf [512];
	snprintf (buf, 512, CONFIG->get_snd_player_args (), sfn.c_str ());
	DL((APP,"Playing \"%s %s\"\n", CONFIG->get_snd_player_cmd (), buf));

	// Better yet to route WAV playing through esd (if it is running):
	// esdplay <fname>.wav
	//
	ASSA::Fork::fork_exec (CONFIG->get_snd_player_cmd (), buf, false);
}

/** 
	Deck information consists of two parts: the header (More... dialog)
    and a list of cards. Either could have been modified by the user.
	In the first case, the status would be MODIFIED_DECK; in the second,
	however, the status would be CLEAN_DECK. We double check here to 
	see if any of the cards have been edited (if new card were added 
	or deleted, the status would be MODIFIED_DECK
*/
DeckEditStatus
Deck::
get_status ()
{	
	if (m_status == CLEAN_DECK) {
		dump_deck_status ();
		cardlist_const_iterator citer = m_cards.begin ();
		while (citer != m_cards.end ()) {
			if ((*citer)->is_dirty ()) {
				m_status = MODIFIED_DECK;
				break;
			}
			citer++;
		}
	}
	dump_deck_status ();
	return m_status;
}

void
Deck::
dump () const 
{
	DL((DECK, "fname = \"%d\"\n", m_fname.c_str ()));
	DL((DECK, "author = \"%s\"\n", m_author.c_str ()));
	DL((DECK, "description = \"%s\"\n", m_desc.c_str ()));
}

void
Deck::
dump_deck_status () const
{
	static const char* names [] = { "NEW_DECK", "INVALID_DECK", 
								   "MODIFIED_DECK", "CLEAN_DECK" };

	DL((APP, "Deck \"%s\"\n", get_name ().c_str ()));
	DL((APP, "Status = <%s>\n", names [m_status]));
}

//------------------------------------------------------------------------------
// LEFTOVERS FOR HEAVY DEBUGGING
//------------------------------------------------------------------------------
#if 0
void
Deck::
traverse_xml_tree (const xmlpp::Node* node_)
{
	trace_with_mask("Deck::traverse_xml_tree",GUITRACE);

	string nodename;
	if (node_) {
		nodename = node_->get_name ();
	}
	const xmlpp::TextNode* node_text;
	node_text = dynamic_cast<const xmlpp::TextNode*> (node_);

	/** Ignore indentation if XML file is indented.
		I wonder how it would work with UTF-8.
	 */
	if (node_text && node_text->is_white_space ()) {
		return;
	}

	DL((DECK,"%sPath = \"%s\"\n", 
		m_indent.c_str(), node_->get_path ().c_str ()));
	
	/** We cast node pointer to all possible types and then 
	 *  determine which one we are dealing with by elimination.
	 */
	const xmlpp::ContentNode* node_content;
	node_content = dynamic_cast<const xmlpp::ContentNode*> (node_);

	const xmlpp::CommentNode* node_comment;
	node_comment = dynamic_cast<const xmlpp::CommentNode*> (node_);

	const xmlpp::Element* node_element;
	node_element = dynamic_cast<const xmlpp::Element*> (node_);

	/** If either Content or Element, indent and print out its name.
	 */
	if (!node_text && !node_comment && !nodename.empty ()) {
		DL((DECK,"%sNode name = %s\n", 
			m_indent.c_str(), nodename.c_str ()));
	}

	/** Find the right type
	 */
	if (node_text) {
		DL((DECK,"%sType = TEXT_NODE\n", m_indent.c_str()));
		DL((DECK,"%stext = \"%s\"\n", m_indent.c_str(), 
			node_text->get_content ().c_str ()));

		if (m_element_node_name == "front_encoding") {
			m_front_encoding = node_text->get_content ();
		}
		else if (m_element_node_name == "back_encoding") {
			m_back_encoding = node_text->get_content ();
		}
		else if (m_element_node_name == "card") {
			m_current_card = new Card;
			m_cards.push_back (m_current_card);
		}
		else if (m_element_node_name == "front") {
			m_current_card->set_question (node_text->get_content ());
		}
		else if (m_element_node_name == "back") {
			m_current_card->set_answer (node_text->get_content ());
		}
		else if (m_element_node_name == "back_example") {
			m_current_card->set_example (node_text->get_content ());
		}
		else if (m_element_node_name == "") {
			= node_text->get_content ();
		}

	}
	else if (node_comment) {
		DL((DECK,"%sType = COMMENT_NODE\n", m_indent.c_str()));
		DL((DECK,"%scomment = \"%s\"\n", m_indent.c_str(), 
			node_comment->get_content ().c_str ()));
	}
	else if (node_content) {
		DL((DECK,"%sType = CONTENT_NODE\n", m_indent.c_str()));
		DL((DECK,"%scontent = \"%s\"\n", m_indent.c_str(), 
			node_content->get_content ().c_str ()));
	}
	else if (node_element) {
		DL((DECK,"%sType = ELEMENT_NODE\n", m_indent.c_str()));
		DL((DECK,"%sline = \"%d\"\n", m_indent.c_str(), 
			node_element->get_line ()));     // line() is for Elements

		/** Print attributes of the element
		 */
		const xmlpp::Element::AttributeList& 
			attributes = node_element->get_attributes ();
		const xmlpp::Attribute* attr;
		xmlpp::Element::AttributeList::const_iterator iter = attributes.begin();
		while (iter != attributes.end ()) {
			attr = *iter;
			DL((DECK,"%sType = ATTRIBUTE_NODE\n", m_indent.c_str()));
			DL((DECK,"%sAttribute: %s = \"%s\"\n", m_indent.c_str(),
				attr->get_name ().c_str (), attr->get_value ().c_str ()));
			iter++;
		}
		attr = node_element->get_attribute ("id");
		if (attr) {
			DL((DECK,"%sID found = %s\n", m_indent.c_str(), 
				attr->get_value().c_str()));
			if (m_element_node_name == "card") {
				std::string value = attr->get_value ();
				m_current_card->set_id (atol (value.substr (1, value.size ())));
				m_element_node_name = "":
			}

		}
	}

	/** If not Content, recurse through the children
	 */
	m_element_node_name = "";
	if (!node_content) {
		xmlpp::Node::NodeList list = node_->get_children ();
		xmlpp::Node::NodeList::iterator iter = list.begin ();
		while (iter != list.end ()) {
			m_indent += "  ";
			traverse_xml_tree (*iter);
			iter++;
		}
	}
	m_indent = "";
}
#endif
